/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "CksClientBoard.H"
#include "CksClientMove.H"
#include <vector>
#include <algorithm> 

void CksClientBoard::setup()
{
  clear();
}

sint4 CksClientBoard::char2cont(char c) const
{
  switch (c) {
  case 'b': return BLACK;
  case 'B': return BLACK_KING;    
  case 'w': return WHITE;
  case 'W': return WHITE_KING;    
  case '-': case ' ': return EMPTY;
  default:
    ERR("illegal char in board.sgf");
  }
  return EMPTY;
}

sint1 CksClientBoard::cont2char(sint4 co) const
{
  switch (co) {
  case BLACK:      return 'b';
  case BLACK_KING: return 'B';    
  case WHITE:      return 'w';
  case WHITE_KING: return 'W';    
  case EMPTY:      return '-';
  case BORDER:     return '#';
  default:
    ERR("illegal cont");
  }
  return '?';
}

bool CksClientBoard::no_moves(bool opponent) const
{
  CksClientBoard bo = *this;
  
  if (opponent) bo.toggle_to_move();

  FORS (i, ClientBoard::MAX_BOARD_SIZE) {
    CksClientMove mv(i);
    if (bo.make_move(mv)) return false;
  }

  return true;
}


int CksClientBoard::disc_num(sint4 cont) const
{
  int n = 0;
  
  FORS (i, ClientBoard::MAX_BOARD_SIZE) {
    if (sq[i] == cont) n++;
  }

  return n;
}

int CksClientBoard::disc_num() const
{
  int n = 0;
  
  FORS (i, ClientBoard::MAX_BOARD_SIZE) {
    if (sq[i] == BLACK || sq[i] == WHITE) n++;
  }

  return n;
}


bool CksClientBoard::moves_available(bool only_jumps) const
{
  sint4 own_checker, own_king;
  sint4 opp_checker, opp_king;
  sint4 dir;
  
  if (turn_color == BLACK) {
    own_checker = BLACK; own_king = BLACK_KING; dir = CksClientBoard::DX;
    opp_checker = WHITE; opp_king = WHITE_KING;
  } else {
    own_checker = WHITE; own_king = WHITE_KING; dir = -CksClientBoard::DX;
    opp_checker = BLACK; opp_king = BLACK_KING;
  }

  FORS (i, CksClientBoard::MAX_BOARD_SIZE) {
    if (sq[i] == own_checker || sq[i] == own_king) {
      
      // checker jump possible?

      if ((sq[i+dir+1] == opp_checker || sq[i+dir+1] == opp_king) &&
	  sq[i+2*(dir+1)] == EMPTY) return true;

      if ((sq[i+dir-1] == opp_checker || sq[i+dir-1] == opp_king) &&
	  sq[i+2*(dir-1)] == EMPTY) return true;

      // checker move possible?
      
      if (!only_jumps && (sq[i+dir+1] == EMPTY ||
			  sq[i+dir-1] == EMPTY)) return true;
      
      if (sq[i] == own_king) {

	// king jump possible?
	
	if ((sq[i-dir+1] == opp_checker || sq[i-dir+1] == opp_king) &&
	    sq[i+2*(-dir+1)] == EMPTY) return true;
	
	if ((sq[i-dir-1] == opp_checker || sq[i-dir-1] == opp_king) &&
	    sq[i+2*(-dir-1)] == EMPTY) return true;

	// king move possible?
      
	if (!only_jumps && (sq[i-dir+1] == EMPTY ||
			    sq[i-dir-1] == EMPTY)) return true;
      }
    }
  }
  return false;
}


bool CksClientBoard::is_finished() const
{
  return !moves_available(false);
}

// true iff move ok

bool CksClientBoard::make_move(const ClientMove &mv0)
{
  sint4 last_sq, w, y;
  
  sint4 dist[4] = { CksClientBoard::DX+1, -(CksClientBoard::DX+1),
		    CksClientBoard::DX-1, -(CksClientBoard::DX-1) };

  const CksClientMove &mv = dynamic_cast<const CksClientMove&>(mv0);
 
  if (!mv.is_valid()) return false;

  CksClientMove::MoveSeq ms;

  mv.get_coords(ms);

  if (ms.size() < 2) return false;
  
  sint4 own_checker, own_king;
  sint4 opp_checker, opp_king;
  sint4 dir;
  
  if (turn_color == BLACK) {

    own_checker = BLACK; own_king = BLACK_KING; dir = CksClientBoard::DX;
    opp_checker = WHITE; opp_king = WHITE_KING;
    
  } else {

    own_checker = WHITE; own_king = WHITE_KING; dir = -CksClientBoard::DX;
    opp_checker = BLACK; opp_king = BLACK_KING;
    
  }

  if (sq[ms[0]] != own_checker &&
      sq[ms[0]] != own_king) return false;

  assert(moves_available());

  CksClientBoard copy = *this;
  
  bool must_jump = moves_available(true);

  if (must_jump) {

    // move must be sequence of jumps

    for (uint4 i=1; i < ms.size(); i++) {

      if (sq[ms[i]] != EMPTY) goto error;

      if (sq[ms[0]] == own_king) {

	sint4 j;
	
	FOR (j, 4)
	  if (ms[i] == ms[i-1] + 2*dist[j]) break;

	if (j >= 4) goto error;
	
	if (sq[ms[i-1]+dist[j]] != opp_checker &&
	    sq[ms[i-1]+dist[j]] != opp_king) goto error;	

      } else {

	sint4 e = 0;
	
	if      (ms[i] == ms[i-1] + 2*(dir+1)) e = dir+1;
	else if (ms[i] == ms[i-1] + 2*(dir-1)) e = dir-1;

	if (!e) goto error;
	
	if (sq[ms[i-1]+e] != opp_checker &&
	    sq[ms[i-1]+e] != opp_king) goto error;	
      }

      sq[(ms[i]+ms[i-1])/2] = EMPTY;
      sq[ms[i]] = sq[ms[i-1]];
      sq[ms[i-1]] = EMPTY;
    }

    // jumps OK
    
  } else {

    // no jump

    if (ms.size() != 2) goto error;
    if (sq[ms[1]] != EMPTY) goto error;
    
    if (sq[ms[0]] == own_king) {

      if (ms[1] != ms[0]+dir+1 && ms[1] != ms[0]+dir-1 &&
	  ms[1] != ms[0]-dir+1 && ms[1] != ms[0]-dir-1)
	goto error;

    } else {
    
      if (ms[1] != ms[0]+dir+1 && ms[1] != ms[0]+dir-1)
	goto error;
    }

    sq[ms[1]] = sq[ms[0]];
    sq[ms[0]] = EMPTY;
  }

  // Promotion
  
  last_sq = ms.back();
  w = type->board_width();
  y = ClientBoard::ind2y(last_sq);
  
  if (turn_color == ClientColor::BLACK && y == w-1) sq[last_sq] = BLACK_KING;
  if (turn_color == ClientColor::WHITE && y == 0)   sq[last_sq] = WHITE_KING;
  
  toggle_to_move();
  return true;

 error:;

  *this = copy;
  return false;
}

